@@ -20,6 +20,7 @@ import re
 import subprocess
 import sys
 import time
+import operator
 
 import gi
 gi.require_version('Gtk', '3.0')
@@ -38,7 +39,7 @@ from gi.repository import Gdk
 
 from . import actions
 from ..dot.lexer import ParseError
-from ..dot.parser import XDotParser
+from ._xdotparser import XDotParser
 from . import animation
 from . import actions
 from .elements import Graph
@@ -49,12 +50,13 @@ class DotWidget(Gtk.DrawingArea):
 
     # TODO GTK3: Second argument has to be of type Gdk.EventButton instead of object.
     __gsignals__ = {
-        'clicked': (GObject.SIGNAL_RUN_LAST, None, (str, object)),
-        'error': (GObject.SIGNAL_RUN_LAST, None, (str,)),
-        'history': (GObject.SIGNAL_RUN_LAST, None, (bool, bool))
+        'clicked': (GObject.SignalFlags.RUN_LAST, None, (str, object)),
+        'error': (GObject.SignalFlags.RUN_LAST, None, (str,)),
+        'history': (GObject.SignalFlags.RUN_LAST, None, (bool, bool))
     }
 
     filter = 'dot'
+    graphviz_version = None
 
     def __init__(self):
         Gtk.DrawingArea.__init__(self)
@@ -72,13 +74,15 @@ class DotWidget(Gtk.DrawingArea):
         self.add_events(Gdk.EventMask.POINTER_MOTION_MASK |
                         Gdk.EventMask.POINTER_MOTION_HINT_MASK |
                         Gdk.EventMask.BUTTON_RELEASE_MASK |
-                        Gdk.EventMask.SCROLL_MASK)
+                        Gdk.EventMask.SCROLL_MASK |
+                        Gdk.EventMask.SMOOTH_SCROLL_MASK)
         self.connect("motion-notify-event", self.on_area_motion_notify)
         self.connect("scroll-event", self.on_area_scroll_event)
         self.connect("size-allocate", self.on_area_size_allocate)
 
         self.connect('key-press-event', self.on_key_press_event)
         self.last_mtime = None
+        self.mtime_changed = False
 
         GLib.timeout_add(1000, self.update)
 
@@ -98,6 +102,7 @@ class DotWidget(Gtk.DrawingArea):
 
     def set_filter(self, filter):
         self.filter = filter
+        self.graphviz_version = None
 
     def run_filter(self, dotcode):
         if not self.filter:
@@ -122,13 +127,13 @@ class DotWidget(Gtk.DrawingArea):
             sys.stderr.write(error + '\n')
         if p.returncode != 0:
             self.error_dialog(error)
-            return None
         return xdotcode
 
     def _set_dotcode(self, dotcode, filename=None, center=True):
         # By default DOT language is UTF-8, but it accepts other encodings
         assert isinstance(dotcode, bytes)
         xdotcode = self.run_filter(dotcode)
+            
         if xdotcode is None:
             return False
         try:
@@ -146,12 +151,21 @@ class DotWidget(Gtk.DrawingArea):
                 self.last_mtime = None
             else:
                 self.last_mtime = os.stat(filename).st_mtime
+            self.mtime_changed = False
             self.openfilename = filename
             return True
 
     def set_xdotcode(self, xdotcode, center=True):
         assert isinstance(xdotcode, bytes)
-        parser = XDotParser(xdotcode)
+
+        if self.graphviz_version is None and self.filter is not None:
+            stdout = subprocess.check_output([self.filter, '-V'], stderr=subprocess.STDOUT)
+            stdout = stdout.rstrip()
+            mo = re.match(br'^.* - .* version (?P<version>.*) \(.*\)$', stdout)
+            assert mo
+            self.graphviz_version = mo.group('version').decode('ascii')
+
+        parser = XDotParser(xdotcode, graphviz_version=self.graphviz_version)
         self.graph = parser.parse()
         self.zoom_image(self.zoom_ratio, center=center)
 
@@ -168,9 +182,15 @@ class DotWidget(Gtk.DrawingArea):
 
     def update(self):
         if self.openfilename is not None:
-            current_mtime = os.stat(self.openfilename).st_mtime
+            try:
+                current_mtime = os.stat(self.openfilename).st_mtime
+            except OSError:
+                return True
             if current_mtime != self.last_mtime:
                 self.last_mtime = current_mtime
+                self.mtime_changed = True
+            elif self.mtime_changed:
+                self.mtime_changed = False
                 self.reload()
         return True
 
@@ -333,6 +353,15 @@ class DotWidget(Gtk.DrawingArea):
         if event.keyval == Gdk.KEY_p:
             self.on_print()
             return True
+        if event.keyval == Gdk.KEY_t:
+            # toggle toolbar visibility
+            win = widget.get_toplevel()
+            toolbar = win.uimanager.get_widget("/ToolBar")
+            toolbar.set_visible(not toolbar.get_visible())
+            return True
+        if event.keyval == Gdk.KEY_w:
+            self.zoom_to_fit()
+            return True
         return False
 
     print_settings = None
@@ -414,7 +443,8 @@ class DotWidget(Gtk.DrawingArea):
                 if url is not None:
                     self.emit('clicked', url.url, event)
                 else:
-                    jump = self.get_jump(x, y)
+                    ctrl_held = event.state & Gdk.ModifierType.CONTROL_MASK
+                    jump = self.get_jump(x, y, to_dst=ctrl_held)
                     if jump is not None:
                         self.animate_to(jump.x, jump.y)
 
@@ -429,9 +459,13 @@ class DotWidget(Gtk.DrawingArea):
             self.zoom_image(self.zoom_ratio * self.ZOOM_INCREMENT,
                             pos=(event.x, event.y))
             return True
-        if event.direction == Gdk.ScrollDirection.DOWN:
+        elif event.direction == Gdk.ScrollDirection.DOWN:
             self.zoom_image(self.zoom_ratio / self.ZOOM_INCREMENT,
                             pos=(event.x, event.y))
+        else:
+            deltas = event.get_scroll_deltas()
+            self.zoom_image(self.zoom_ratio * (1 - deltas.delta_y / 10),
+                            pos=(event.x, event.y))
             return True
         return False
 
@@ -495,9 +529,9 @@ class DotWidget(Gtk.DrawingArea):
         x, y = self.window2graph(x, y)
         return self.graph.get_url(x, y)
 
-    def get_jump(self, x, y):
+    def get_jump(self, x, y, to_dst = False):
         x, y = self.window2graph(x, y)
-        return self.graph.get_jump(x, y)
+        return self.graph.get_jump(x, y, to_dst)
 
 
 class FindMenuToolAction(Gtk.Action):
@@ -513,6 +547,7 @@ class DotWindow(Gtk.Window):
     <ui>
         <toolbar name="ToolBar">
             <toolitem action="Open"/>
+            <toolitem action="Export"/>
             <toolitem action="Reload"/>
             <toolitem action="Print"/>
             <separator/>
@@ -525,6 +560,10 @@ class DotWindow(Gtk.Window):
             <toolitem action="Zoom100"/>
             <separator/>
             <toolitem name="Find" action="Find"/>
+            <separator name="FindNextSeparator"/>
+            <toolitem action="FindNext"/>
+            <separator name="FindStatusSeparator"/>
+            <toolitem name="FindStatus" action="FindStatus"/>
         </toolbar>
     </ui>
     '''
@@ -540,6 +579,7 @@ class DotWindow(Gtk.Window):
 
         window.set_title(self.base_title)
         window.set_default_size(width, height)
+        window.set_wmclass("xdot", "xdot")
         vbox = Gtk.VBox()
         window.add(vbox)
 
@@ -560,14 +600,16 @@ class DotWindow(Gtk.Window):
 
         # Create actions
         actiongroup.add_actions((
-            ('Open', Gtk.STOCK_OPEN, None, None, None, self.on_open),
-            ('Reload', Gtk.STOCK_REFRESH, None, None, None, self.on_reload),
+            ('Open', Gtk.STOCK_OPEN, None, None, "Open dot-file", self.on_open),
+            ('Export', Gtk.STOCK_SAVE_AS, None, None, "Export graph to other format", self.on_export),
+            ('Reload', Gtk.STOCK_REFRESH, None, None, "Reload graph", self.on_reload),
             ('Print', Gtk.STOCK_PRINT, None, None,
              "Prints the currently visible part of the graph", self.dotwidget.on_print),
-            ('ZoomIn', Gtk.STOCK_ZOOM_IN, None, None, None, self.dotwidget.on_zoom_in),
-            ('ZoomOut', Gtk.STOCK_ZOOM_OUT, None, None, None, self.dotwidget.on_zoom_out),
-            ('ZoomFit', Gtk.STOCK_ZOOM_FIT, None, None, None, self.dotwidget.on_zoom_fit),
-            ('Zoom100', Gtk.STOCK_ZOOM_100, None, None, None, self.dotwidget.on_zoom_100),
+            ('ZoomIn', Gtk.STOCK_ZOOM_IN, None, None, "Zoom in", self.dotwidget.on_zoom_in),
+            ('ZoomOut', Gtk.STOCK_ZOOM_OUT, None, None, "Zoom out", self.dotwidget.on_zoom_out),
+            ('ZoomFit', Gtk.STOCK_ZOOM_FIT, None, None, "Fit zoom", self.dotwidget.on_zoom_fit),
+            ('Zoom100', Gtk.STOCK_ZOOM_100, None, None, "Reset zoom level", self.dotwidget.on_zoom_100),
+            ('FindNext', Gtk.STOCK_GO_FORWARD, 'Next Result', None, 'Move to the next search result', self.on_find_next),
         ))
 
         self.back_action = Gtk.Action('Back', None, None, Gtk.STOCK_GO_BACK)
@@ -584,6 +626,10 @@ class DotWindow(Gtk.Window):
                                          "Find a node by name", None)
         actiongroup.add_action(find_action)
 
+        findstatus_action = FindMenuToolAction("FindStatus", None,
+                                               "Number of results found", None)
+        actiongroup.add_action(findstatus_action)
+
         # Add the actiongroup to the uimanager
         uimanager.insert_action_group(actiongroup, 0)
 
@@ -602,7 +648,7 @@ class DotWindow(Gtk.Window):
 
         # Add Find text search
         find_toolitem = uimanager.get_widget('/ToolBar/Find')
-        self.textentry = Gtk.Entry(max_length=20)
+        self.textentry = Gtk.Entry()
         self.textentry.set_icon_from_stock(0, Gtk.STOCK_FIND)
         find_toolitem.add(self.textentry)
 
@@ -610,18 +656,34 @@ class DotWindow(Gtk.Window):
         self.textentry.connect("activate", self.textentry_activate, self.textentry);
         self.textentry.connect("changed", self.textentry_changed, self.textentry);
 
+        uimanager.get_widget('/ToolBar/FindNextSeparator').set_draw(False)
+        uimanager.get_widget('/ToolBar/FindStatusSeparator').set_draw(False)
+        self.find_next_toolitem = uimanager.get_widget('/ToolBar/FindNext')
+        self.find_next_toolitem.set_sensitive(False)
+
+        self.find_count = Gtk.Label()
+        findstatus_toolitem = uimanager.get_widget('/ToolBar/FindStatus')
+        findstatus_toolitem.add(self.find_count)
+
         self.show_all()
 
     def find_text(self, entry_text):
         found_items = []
         dot_widget = self.dotwidget
-        regexp = re.compile(entry_text)
-        for element in dot_widget.graph.nodes + dot_widget.graph.edges:
+        try:
+            regexp = re.compile(entry_text)
+        except re.error as err:
+            sys.stderr.write('warning: re.compile() failed with error "%s"\n' % err)
+            return []
+        for element in dot_widget.graph.nodes + dot_widget.graph.edges + dot_widget.graph.shapes:
             if element.search_text(regexp):
                 found_items.append(element)
-        return found_items
+        return sorted(found_items, key=operator.methodcaller('get_text'))
 
     def textentry_changed(self, widget, entry):
+        self.find_count.set_label('')
+        self.find_index = 0
+        self.find_next_toolitem.set_sensitive(False)
         entry_text = entry.get_text()
         dot_widget = self.dotwidget
         if not entry_text:
@@ -630,18 +692,24 @@ class DotWindow(Gtk.Window):
 
         found_items = self.find_text(entry_text)
         dot_widget.set_highlight(found_items, search=True)
+        if found_items:
+            self.find_count.set_label('%d nodes found' % len(found_items))
 
     def textentry_activate(self, widget, entry):
+        self.find_index = 0
+        self.find_next_toolitem.set_sensitive(False)
         entry_text = entry.get_text()
         dot_widget = self.dotwidget
         if not entry_text:
             dot_widget.set_highlight(None, search=True)
+            self.set_focus(self.dotwidget)
             return
 
         found_items = self.find_text(entry_text)
         dot_widget.set_highlight(found_items, search=True)
-        if(len(found_items) == 1):
+        if found_items:
             dot_widget.animate_to(found_items[0].x, found_items[0].y)
+        self.find_next_toolitem.set_sensitive(len(found_items) > 1)
 
     def set_filter(self, filter):
         self.dotwidget.set_filter(filter)
@@ -672,7 +740,7 @@ class DotWindow(Gtk.Window):
 
     def on_open(self, action):
         chooser = Gtk.FileChooserDialog(parent=self,
-                                        title="Open dot File",
+                                        title="Open Graphviz File",
                                         action=Gtk.FileChooserAction.OPEN,
                                         buttons=(Gtk.STOCK_CANCEL,
                                                  Gtk.ResponseType.CANCEL,
@@ -681,7 +749,8 @@ class DotWindow(Gtk.Window):
         chooser.set_default_response(Gtk.ResponseType.OK)
         chooser.set_current_folder(self.last_open_dir)
         filter = Gtk.FileFilter()
-        filter.set_name("Graphviz dot files")
+        filter.set_name("Graphviz files")
+        filter.add_pattern("*.gv")
         filter.add_pattern("*.dot")
         chooser.add_filter(filter)
         filter = Gtk.FileFilter()
@@ -695,6 +764,68 @@ class DotWindow(Gtk.Window):
             self.open_file(filename)
         else:
             chooser.destroy()
+   
+    def export_file(self, filename, format_):
+        if not filename.endswith("." + format_):
+            filename += '.' + format_
+        cmd = [
+            self.dotwidget.filter, # program name, usually "dot"
+            '-T' + format_,
+            '-o', filename,
+            self.dotwidget.openfilename,
+        ]
+        subprocess.check_call(cmd)
+
+    def on_export(self, action):
+        
+        if self.dotwidget.openfilename is None:
+            return
+        
+        default_filter = "PNG image"
+    
+        output_formats = {
+            "dot file": "dot",
+            "GIF image": "gif",
+            "JPG image": "jpg",
+            "JSON": "json",
+            "PDF": "pdf",
+            "PNG image": "png",
+            "PostScript": "ps",
+            "SVG image": "svg",
+            "XFIG image": "fig",
+            "xdot file": "xdot",
+        }
+        buttons = (
+            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
+            Gtk.STOCK_SAVE, Gtk.ResponseType.OK)
+        chooser = Gtk.FileChooserDialog(
+            parent=self,
+            title="Export to other file format.",
+            action=Gtk.FileChooserAction.SAVE,
+            buttons=buttons) 
+        chooser.set_default_response(Gtk.ResponseType.OK)
+        chooser.set_current_folder(self.last_open_dir)
+        
+        openfilename = os.path.basename(self.dotwidget.openfilename)
+        openfileroot = os.path.splitext(openfilename)[0]
+        chooser.set_current_name(openfileroot)
+
+        for name, ext in output_formats.items():
+            filter_ = Gtk.FileFilter()
+            filter_.set_name(name)
+            filter_.add_pattern('*.' + ext)
+            chooser.add_filter(filter_)
+            if name == default_filter:
+                chooser.set_filter(filter_)
+
+        if chooser.run() == Gtk.ResponseType.OK:
+            filename = chooser.get_filename()
+            format_ = output_formats[chooser.get_filter().get_name()]
+            chooser.destroy()
+            self.export_file(filename, format_)
+        else:
+            chooser.destroy()
+	
 
     def on_reload(self, action):
         self.dotwidget.reload()
@@ -708,6 +839,15 @@ class DotWindow(Gtk.Window):
         dlg.run()
         dlg.destroy()
 
+    def on_find_next(self, action):
+        self.find_index += 1
+        entry_text = self.textentry.get_text()
+        # Maybe storing the search result would be better
+        found_items = self.find_text(entry_text)
+        found_item = found_items[self.find_index]
+        self.dotwidget.animate_to(found_item.x, found_item.y)
+        self.find_next_toolitem.set_sensitive(len(found_items) > self.find_index + 1)
+
     def on_history(self, action, has_back, has_forward):
         self.back_action.set_sensitive(has_back)
         self.forward_action.set_sensitive(has_forward)
